@@ -96,11 +96,14 @@ static VALUE half_in_rational;
 static ID idPow, idPLUS, idMINUS, idSTAR, idDIV, idGE;
 static ID id_eqeq_p, id_idiv, id_negate, id_to_f, id_cmp, id_nan_p;
 static ID id_each, id_real_p, id_sum, id_population, id_closed, id_edge;
+static ID id_skip_na;
 
-static VALUE sym_auto, sym_left, sym_right;
+static VALUE sym_auto, sym_left, sym_right, sym_sturges;
 
 static VALUE cHistogram;
 
+static VALUE orig_enum_sum, orig_ary_sum;
+
 inline static VALUE
 f_add(VALUE x, VALUE y)
 {
@@ -135,7 +138,7 @@ complex_new(VALUE klass, VALUE real, VALUE imag)
 {
   assert(!RB_TYPE_P(real, T_COMPLEX));
 
-  NEWOBJ_OF(obj, struct RComplex, klass, T_COMPLEX | (RGENGC_WB_PROTECTED_COMPLEX ? FL_WB_PROTECTED : 0));
+  VALUE obj = rb_get_alloc_func(klass)(klass);
 
   RCOMPLEX_SET_REAL(obj, real);
   RCOMPLEX_SET_IMAG(obj, imag);
@@ -535,9 +538,11 @@ f_gcd(VALUE x, VALUE y)
 inline static VALUE
 nurat_s_new_internal(VALUE klass, VALUE num, VALUE den)
 {
-  NEWOBJ_OF(obj, struct RRational, klass, T_RATIONAL | (RGENGC_WB_PROTECTED_RATIONAL ? FL_WB_PROTECTED : 0));
+  VALUE obj = rb_get_alloc_func(klass)(klass);
+
   RRATIONAL_SET_NUM(obj, num);
   RRATIONAL_SET_DEN(obj, den);
+
   return (VALUE)obj;
 }
 
@@ -632,39 +637,67 @@ rb_rational_plus(VALUE self, VALUE other)
 }
 #endif
 
-/* call-seq:
- *    ary.sum
- *
- * Calculate the sum of the values in `ary`.
- * This method utilizes
- * [Kahan summation algorithm](https://en.wikipedia.org/wiki/Kahan_summation_algorithm)
- * to compensate the result precision when the `ary` includes Float values.
- *
- * Note that This library does not redefine `sum` method introduced in Ruby 2.4.
- *
- * @return [Number] A summation value
- */
-static VALUE
-ary_sum(int argc, VALUE* argv, VALUE ary)
+static inline int
+is_na(VALUE v)
+{
+  if (NIL_P(v))
+    return 1;
+
+  if (RB_FLOAT_TYPE_P(v) && isnan(RFLOAT_VALUE(v)))
+    return 1;
+
+  if (rb_respond_to(v, id_nan_p) && RTEST(rb_funcall(v, id_nan_p, 0)))
+    return 1;
+
+  return 0;
+}
+
+static int opt_skip_na(VALUE opts)
+{
+  VALUE skip_na = Qfalse;
+
+  if (!NIL_P(opts)) {
+#ifdef HAVE_RB_GET_KWARGS
+    ID kwargs = id_skip_na;
+    rb_get_kwargs(opts, &kwargs, 0, 1, &skip_na);
+#else
+    VALUE val = rb_hash_aref(opts, ID2SYM(id_skip_na));
+    skip_na = NIL_P(val) ? skip_na : val;
+#endif
+  }
+
+  return RTEST(skip_na);
+}
+
+VALUE
+ary_calculate_sum(VALUE ary, VALUE init, int skip_na, long *na_count_out)
 {
   VALUE e, v, r;
   long i, n;
   int block_given;
-
-  if (rb_scan_args(argc, argv, "01", &v) == 0)
-    v = LONG2FIX(0);
+  long na_count = 0;
 
   block_given = rb_block_given_p();
 
-  if (RARRAY_LEN(ary) == 0)
-    return v;
+  if (RARRAY_LEN(ary) == 0) {
+    if (na_count_out != NULL) {
+      *na_count_out = 0;
+    }
+    return init;
+  }
 
   n = 0;
   r = Qundef;
+  v = init;
   for (i = 0; i < RARRAY_LEN(ary); i++) {
     e = RARRAY_AREF(ary, i);
     if (block_given)
       e = rb_yield(e);
+    if (skip_na && is_na(e)) {
+      ++na_count;
+      continue;
+    }
+
     if (FIXNUM_P(e)) {
       n += FIX2LONG(e); /* should not overflow long type */
       if (!FIXABLE(n)) {
@@ -688,7 +721,7 @@ ary_sum(int argc, VALUE* argv, VALUE ary)
     v = rb_fix_plus(LONG2FIX(n), v);
   if (r != Qundef)
     v = rb_rational_plus(r, v);
-  return v;
+  goto finish;
 
 not_exact:
   if (n != 0)
@@ -708,6 +741,11 @@ not_exact:
       e = RARRAY_AREF(ary, i);
       if (block_given)
         e = rb_yield(e);
+      if (skip_na && is_na(e)) {
+        ++na_count;
+        continue;
+      }
+
       if (RB_FLOAT_TYPE_P(e))
         has_float_value:
           x = RFLOAT_VALUE(e);
@@ -725,7 +763,9 @@ not_exact:
       c = (t - f) - y;
       f = t;
     }
-    return DBL2NUM(f);
+
+    v = DBL2NUM(f);
+    goto finish;
 
   not_float:
     v = DBL2NUM(f);
@@ -736,13 +776,53 @@ not_exact:
     e = RARRAY_AREF(ary, i);
     if (block_given)
       e = rb_yield(e);
+    if (skip_na && is_na(e)) {
+      ++na_count;
+      continue;
+    }
   has_some_value:
     v = rb_funcall(v, idPLUS, 1, e);
   }
 
+finish:
+  if (na_count_out != NULL) {
+    *na_count_out = na_count;
+  }
   return v;
 }
 
+/* call-seq:
+ *    ary.sum(skip_na: false)
+ *
+ * Calculate the sum of the values in `ary`.
+ * This method utilizes
+ * [Kahan summation algorithm](https://en.wikipedia.org/wiki/Kahan_summation_algorithm)
+ * to compensate the result precision when the `ary` includes Float values.
+ *
+ * Note that This library does not redefine `sum` method introduced in Ruby 2.4.
+ *
+ * @return [Number] A summation value
+ */
+static VALUE
+ary_sum(int argc, VALUE* argv, VALUE ary)
+{
+  VALUE v, opts;
+  int skip_na;
+
+  if (rb_scan_args(argc, argv, "01:", &v, &opts) == 0) {
+    v = LONG2FIX(0);
+  }
+  skip_na = opt_skip_na(opts);
+
+#ifndef HAVE_ENUM_SUM
+  if (!skip_na) {
+    return rb_funcall(orig_ary_sum, rb_intern("call"), argc, &v);
+  }
+#endif
+
+  return ary_calculate_sum(ary, v, skip_na, NULL);
+}
+
 static void
 calculate_and_set_mean(VALUE *mean_ptr, VALUE sum, long const n)
 {
@@ -771,9 +851,10 @@ calculate_and_set_mean(VALUE *mean_ptr, VALUE sum, long const n)
 }
 
 static void
-ary_mean_variance(VALUE ary, VALUE *mean_ptr, VALUE *variance_ptr, size_t ddof)
+ary_mean_variance(VALUE ary, VALUE *mean_ptr, VALUE *variance_ptr, size_t ddof, int skip_na)
 {
   long i;
+  long na_count;
   size_t n = 0;
   double m = 0.0, m2 = 0.0, f = 0.0, c = 0.0;
 
@@ -797,8 +878,8 @@ ary_mean_variance(VALUE ary, VALUE *mean_ptr, VALUE *variance_ptr, size_t ddof)
 
   if (variance_ptr == NULL) {
     VALUE init = DBL2NUM(0.0);
-    VALUE const sum = ary_sum(1, &init, ary);
-    long const n = RARRAY_LEN(ary);
+    VALUE const sum = ary_calculate_sum(ary, init, skip_na, &na_count);
+    long const n = RARRAY_LEN(ary) - na_count;
     calculate_and_set_mean(mean_ptr, sum, n);
     return;
   }
@@ -839,26 +920,46 @@ ary_mean_variance(VALUE ary, VALUE *mean_ptr, VALUE *variance_ptr, size_t ddof)
   }
 }
 
-static int
-opt_population_p(VALUE opts)
+struct variance_opts {
+  int population;
+  int skip_na;
+};
+
+static void
+get_variance_opts(VALUE opts, struct variance_opts *out)
 {
-  VALUE population = Qfalse;
+  assert(out != NULL);
+
+  out->population = 0;
+  out->skip_na = 0;
 
   if (!NIL_P(opts)) {
 #ifdef HAVE_RB_GET_KWARGS
-    ID kwargs = id_population;
-    rb_get_kwargs(opts, &kwargs, 0, 1, &population);
+    static ID kwarg_keys[2];
+    VALUE kwarg_vals;
+
+    if (!kwarg_keys[0]) {
+      kwarg_keys[0] = id_population;
+      kwarg_keys[1] = id_skip_na;
+    }
+
+    rb_get_kwargs(opts, &kwarg_keys, 0, 2, kwarg_vals);
+    out->population = (kwarg_vals[0] != Qundef) ? RTEST(kwarg_vals[0]) : out->population;
+    out->skip_na = (kwarg_vals[1] != Qundef) ? RTEST(kwarg_vals[1]) : out->skip_na;
 #else
-    VALUE val = rb_hash_aref(opts, ID2SYM(id_population));
-    population = NIL_P(val) ? population : val;
+    VALUE val;
+
+    val = rb_hash_aref(opts, ID2SYM(id_population));
+    out->population = NIL_P(val) ? out->population : RTEST(val);
+
+    val = rb_hash_aref(opts, ID2SYM(id_skip_na));
+    out->skip_na = NIL_P(val) ? out->skip_na : RTEST(val);
 #endif
   }
-
-  return RTEST(population);
 }
 
 /* call-seq:
- *    ary.mean_variance(population: false)
+ *    ary.mean_variance(population: false, skip_na: false)
  *
  * Calculate a mean and a variance of the values in `ary`.
  * The first element of the result array is the mean, and the second is the variance.
@@ -876,19 +977,21 @@ opt_population_p(VALUE opts)
 static VALUE
 ary_mean_variance_m(int argc, VALUE* argv, VALUE ary)
 {
-  VALUE opts, mean, variance;
+  struct variance_opts options;
+  VALUE opts, mean = Qnil, variance = Qnil;
   size_t ddof = 1;
 
   rb_scan_args(argc, argv, "0:", &opts);
-  if (opt_population_p(opts))
+  get_variance_opts(opts, &options);
+  if (options.population)
     ddof = 0;
 
-  ary_mean_variance(ary, &mean, &variance, ddof);
+  ary_mean_variance(ary, &mean, &variance, ddof, options.skip_na);
   return rb_assoc_new(mean, variance);
 }
 
 /* call-seq:
- *    ary.mean
+ *    ary.mean(skip_na: false)
  *
  * Calculate a mean of the values in `ary`.
  * This method utilizes
@@ -898,15 +1001,20 @@ ary_mean_variance_m(int argc, VALUE* argv, VALUE ary)
  * @return [Number] A mean value
  */
 static VALUE
-ary_mean(VALUE ary)
+ary_mean(int argc, VALUE *argv, VALUE ary)
 {
-  VALUE mean;
-  ary_mean_variance(ary, &mean, NULL, 1);
+  VALUE mean = Qnil, opts;
+  int skip_na;
+
+  rb_scan_args(argc, argv, ":", &opts);
+  skip_na = opt_skip_na(opts);
+
+  ary_mean_variance(ary, &mean, NULL, 1, skip_na);
   return mean;
 }
 
 /* call-seq:
- *    ary.variance(population: false)
+ *    ary.variance(population: false, skip_na: false)
  *
  * Calculate a variance of the values in `ary`.
  * This method scan values in `ary` only once,
@@ -922,14 +1030,16 @@ ary_mean(VALUE ary)
 static VALUE
 ary_variance(int argc, VALUE* argv, VALUE ary)
 {
+  struct variance_opts options;
   VALUE opts, variance;
   size_t ddof = 1;
 
   rb_scan_args(argc, argv, "0:", &opts);
-  if (opt_population_p(opts))
+  get_variance_opts(opts, &options);
+  if (options.population)
     ddof = 0;
 
-  ary_mean_variance(ary, NULL, &variance, ddof);
+  ary_mean_variance(ary, NULL, &variance, ddof, options.skip_na);
   return variance;
 }
 
@@ -943,6 +1053,7 @@ struct enum_sum_memo {
   double f, c;
   int block_given;
   int float_value;
+  int skip_na;
 };
 
 static void
@@ -956,8 +1067,12 @@ sum_iter(VALUE e, struct enum_sum_memo *memo)
   double f = memo->f;
   double c = memo->c;
 
-  if (memo->block_given)
+  if (memo->block_given) {
     e = rb_yield(e);
+  }
+  if (memo->skip_na && is_na(e)) {
+    return;
+  }
 
   memo->count += 1;
 
@@ -1090,7 +1205,7 @@ int_range_sum_count(VALUE beg, VALUE end, int excl,
 }
 
 static void
-enum_sum_count(VALUE obj, VALUE init, VALUE *sum_ptr, long *count_ptr)
+enum_sum_count(VALUE obj, VALUE init, int skip_na, VALUE *sum_ptr, long *count_ptr)
 {
   struct enum_sum_memo memo;
   VALUE beg, end;
@@ -1101,6 +1216,7 @@ enum_sum_count(VALUE obj, VALUE init, VALUE *sum_ptr, long *count_ptr)
   memo.block_given = rb_block_given_p();
   memo.n = 0;
   memo.r = Qundef;
+  memo.skip_na = skip_na;
 
   if ((memo.float_value = RB_FLOAT_TYPE_P(memo.v))) {
     memo.f = RFLOAT_VALUE(memo.v);
@@ -1138,9 +1254,8 @@ enum_sum_count(VALUE obj, VALUE init, VALUE *sum_ptr, long *count_ptr)
     *count_ptr = memo.count;
 }
 
-#ifndef HAVE_ENUM_SUM
 /* call-seq:
- *    enum.sum
+ *    enum.sum(skip_na: false)
  *
  * Calculate the sum of the values in `enum`.
  * This method utilizes
@@ -1154,16 +1269,27 @@ enum_sum_count(VALUE obj, VALUE init, VALUE *sum_ptr, long *count_ptr)
 static VALUE
 enum_sum(int argc, VALUE* argv, VALUE obj)
 {
-  VALUE sum, init;
+  VALUE sum, init, opts;
+  int skip_na;
 
-  if (rb_scan_args(argc, argv, "01", &init) == 0)
+  if (rb_scan_args(argc, argv, "01:", &init, &opts) == 0) {
     init = LONG2FIX(0);
+  }
+  skip_na = opt_skip_na(opts);
 
-  enum_sum_count(obj, init, &sum, NULL);
+#ifndef HAVE_ENUM_SUM
+  if (skip_na) {
+    enum_sum_count(obj, init, skip_na, &sum, NULL);
+  }
+  else {
+    rb_funcall(orig_enum_sum, rb_intern("call"), argc, &init);
+  }
+#else
+  enum_sum_count(obj, init, skip_na, &sum, NULL);
+#endif
 
   return sum;
 }
-#endif
 
 struct enum_mean_variance_memo {
   int block_given;
@@ -1253,7 +1379,7 @@ enum_mean_variance(VALUE obj, VALUE *mean_ptr, VALUE *variance_ptr, size_t ddof)
     long n;
     VALUE sum;
     VALUE init = DBL2NUM(0.0);
-    enum_sum_count(obj, init, &sum, &n);
+    enum_sum_count(obj, init, 0, &sum, &n); /* TODO: skip_na */
     if (n > 0)
       calculate_and_set_mean(mean_ptr, sum, n);
     return;
@@ -1303,11 +1429,13 @@ enum_mean_variance(VALUE obj, VALUE *mean_ptr, VALUE *variance_ptr, size_t ddof)
 static VALUE
 enum_mean_variance_m(int argc, VALUE* argv, VALUE obj)
 {
+  struct variance_opts options;
   VALUE opts, mean, variance;
   size_t ddof = 1;
 
   rb_scan_args(argc, argv, "0:", &opts);
-  if (opt_population_p(opts))
+  get_variance_opts(opts, &options);
+  if (options.population)
     ddof = 0;
 
   enum_mean_variance(obj, &mean, &variance, ddof);
@@ -1349,11 +1477,13 @@ enum_mean(VALUE obj)
 static VALUE
 enum_variance(int argc, VALUE* argv, VALUE obj)
 {
+  struct variance_opts options;
   VALUE opts, variance;
   size_t ddof = 1;
 
   rb_scan_args(argc, argv, "0:", &opts);
-  if (opt_population_p(opts))
+  get_variance_opts(opts, &options);
+  if (options.population)
     ddof = 0;
 
   enum_mean_variance(obj, NULL, &variance, ddof);
@@ -1392,11 +1522,13 @@ sqrt_value(VALUE x)
 static VALUE
 enum_mean_stdev(int argc, VALUE* argv, VALUE obj)
 {
+  struct variance_opts options;
   VALUE opts, mean, variance;
   size_t ddof = 1;
 
   rb_scan_args(argc, argv, "0:", &opts);
-  if (opt_population_p(opts))
+  get_variance_opts(opts, &options);
+  if (options.population)
     ddof = 0;
 
   enum_mean_variance(obj, &mean, &variance, ddof);
@@ -1446,14 +1578,16 @@ enum_stdev(int argc, VALUE* argv, VALUE obj)
 static VALUE
 ary_mean_stdev(int argc, VALUE* argv, VALUE ary)
 {
+  struct variance_opts options;
   VALUE opts, mean, variance;
   size_t ddof = 1;
 
   rb_scan_args(argc, argv, "0:", &opts);
-  if (opt_population_p(opts))
+  get_variance_opts(opts, &options);
+  if (options.population)
     ddof = 0;
 
-  ary_mean_variance(ary, &mean, &variance, ddof);
+  ary_mean_variance(ary, &mean, &variance, ddof, options.skip_na);
   VALUE stdev = sqrt_value(variance);
   return rb_assoc_new(mean, stdev);
 }
@@ -1479,21 +1613,6 @@ ary_stdev(int argc, VALUE* argv, VALUE ary)
   return stdev;
 }
 
-static inline int
-is_na(VALUE v)
-{
-  if (NIL_P(v))
-    return 1;
-
-  if (RB_FLOAT_TYPE_P(v) && isnan(RFLOAT_VALUE(v)))
-    return 1;
-
-  if (rb_respond_to(v, id_nan_p) && RTEST(rb_funcall(v, id_nan_p, 0)))
-    return 1;
-
-  return 0;
-}
-
 static int
 ary_percentile_sort_cmp(const void *ap, const void *bp, void *dummy)
 {
@@ -1900,7 +2019,7 @@ any_value_counts(int argc, VALUE *argv, VALUE obj,
   struct value_counts_opts opts;
   struct value_counts_memo memo;
 
-  rb_scan_args(argc, argv, ":", &kwargs);
+  rb_scan_args(argc, argv, "0:", &kwargs);
   value_counts_extract_opts(kwargs, &opts);
 
   memo.result = rb_hash_new();
@@ -2255,9 +2374,9 @@ ary_histogram_calculate_edge_lo_hi(const double lo, const double hi, const long
 }
 
 static VALUE
-ary_histogram_calculate_edge(VALUE ary, const long nbins, const int left_p)
+ary_histogram_calculate_edge(VALUE ary, VALUE arg0, const int left_p)
 {
-  long n;
+  long n, nbins;
   VALUE minmax;
   VALUE edge = Qnil;
   double lo, hi;
@@ -2265,6 +2384,22 @@ ary_histogram_calculate_edge(VALUE ary, const long nbins, const int left_p)
   Check_Type(ary, T_ARRAY);
   n = RARRAY_LEN(ary);
 
+  if (NIL_P(arg0)) {
+    arg0 = sym_auto;
+  }
+
+  if (RB_TYPE_P(arg0, T_SYMBOL)) {
+    if (arg0 != sym_auto && arg0 != sym_sturges) {
+      rb_raise(rb_eArgError, "Unknown method to calculate bin width: %+"PRIsVALUE, arg0);
+    }
+    else {
+      nbins = sturges(n);
+    }
+  }
+  else {
+    nbins = NUM2LONG(arg0);
+  }
+
   if (n == 0 && nbins < 0) {
     rb_raise(rb_eArgError, "nbins must be >= 0 for an empty array, got %ld", nbins);
   }
@@ -2337,19 +2472,13 @@ static VALUE
 ary_histogram(int argc, VALUE *argv, VALUE ary)
 {
   VALUE arg0, kwargs, bin_weights;
-  long nbins, n_bin_weights, i;
+  long n_bin_weights, i;
 
   VALUE weight_array = Qnil;
   VALUE edges = Qnil;
   int left_p = 1;
 
   rb_scan_args(argc, argv, "01:", &arg0, &kwargs);
-  if (NIL_P(arg0) || arg0 == sym_auto) {
-    nbins = sturges(RARRAY_LEN(ary));
-  }
-  else {
-    nbins = NUM2LONG(arg0);
-  }
 
   if (!NIL_P(kwargs)) {
     enum { kw_weights, kw_edges, kw_closed };
@@ -2370,7 +2499,10 @@ ary_histogram(int argc, VALUE *argv, VALUE ary)
   }
 
   if (NIL_P(edges)) {
-    edges = ary_histogram_calculate_edge(ary, nbins, left_p);
+    edges = ary_histogram_calculate_edge(ary, arg0, left_p);
+  }
+  else if (! NIL_P(arg0)) {
+    rb_raise(rb_eArgError, "Unable to use both `nbins` and `edges` together");
   }
 
   n_bin_weights = RARRAY_LEN(edges) - 1;
@@ -2395,10 +2527,12 @@ Init_extension(void)
   rb_ext_ractor_safe(true);
 #endif
 
-#ifndef HAVE_ENUM_SUM
-  rb_define_method(rb_mEnumerable, "sum", enum_sum, -1);
-#endif
+  mEnumerableStatistics = rb_const_get_at(rb_cObject, rb_intern("EnumerableStatistics"));
 
+  orig_enum_sum = rb_funcall(rb_mEnumerable, rb_intern("public_instance_method"), 1, rb_str_new_cstr("sum"));
+  orig_ary_sum = rb_funcall(rb_cArray, rb_intern("public_instance_method"), 1, rb_str_new_cstr("sum"));
+
+  rb_define_method(rb_mEnumerable, "sum", enum_sum, -1);
   rb_define_method(rb_mEnumerable, "mean_variance", enum_mean_variance_m, -1);
   rb_define_method(rb_mEnumerable, "mean", enum_mean, 0);
   rb_define_method(rb_mEnumerable, "variance", enum_variance, -1);
@@ -2406,11 +2540,9 @@ Init_extension(void)
   rb_define_method(rb_mEnumerable, "stdev", enum_stdev, -1);
   rb_define_method(rb_mEnumerable, "value_counts", enum_value_counts, -1);
 
-#ifndef HAVE_ARRAY_SUM
   rb_define_method(rb_cArray, "sum", ary_sum, -1);
-#endif
   rb_define_method(rb_cArray, "mean_variance", ary_mean_variance_m, -1);
-  rb_define_method(rb_cArray, "mean", ary_mean, 0);
+  rb_define_method(rb_cArray, "mean", ary_mean, -1);
   rb_define_method(rb_cArray, "variance", ary_variance, -1);
   rb_define_method(rb_cArray, "mean_stdev", ary_mean_stdev, -1);
   rb_define_method(rb_cArray, "stdev", ary_stdev, -1);
@@ -2423,7 +2555,6 @@ Init_extension(void)
   half_in_rational = nurat_s_new_internal(rb_cRational, INT2FIX(1), INT2FIX(2));
   rb_gc_register_mark_object(half_in_rational);
 
-  mEnumerableStatistics = rb_const_get_at(rb_cObject, rb_intern("EnumerableStatistics"));
   cHistogram = rb_const_get_at(mEnumerableStatistics, rb_intern("Histogram"));
 
   rb_define_method(rb_cArray, "histogram", ary_histogram, -1);
@@ -2449,6 +2580,7 @@ Init_extension(void)
   id_population = rb_intern("population");
   id_closed = rb_intern("closed");
   id_edge = rb_intern("edge");
+  id_skip_na = rb_intern("skip_na");
 
   sym_auto = ID2SYM(rb_intern("auto"));
   sym_left = ID2SYM(rb_intern("left"));
